from contextlib import contextmanager
from datetime import datetime
import json
import logging
import os
import tempfile
from typing import Dict, List, Union, Tuple, Optional, Type, cast
from uuid import uuid4

from azureml.core import Workspace
from azureml.core.authentication import (
    InteractiveLoginAuthentication,
    ServicePrincipalAuthentication,
)
import mlflow
from mlflow.entities import Metric, Param
from mlflow.tracking import MlflowClient
from pytz import UTC

from gordo.builder import ModelBuilder
from gordo.machine import Machine
from gordo.machine.encoders import MachineJSONEncoder
from gordo.util.utils import capture_args
from gordo_core.sensor_tag import extract_tag_name
from gordo.builder.utils import create_model_builder

from .base import BaseReporter
from .exceptions import ReporterException

logger = logging.getLogger(__name__)


class MlflowLoggingError(ReporterException):
    pass


def _validate_dict(d: dict, required_keys: List[str]):
    """
    Validate the required keys are contained in provided dictionary

    Parameters
    ----------
    d: dict
        Dictionary to validate.
    required_keys: List[str]
        Keys that must be present in provided dictionary.
    """
    if any([key not in d for key in required_keys]):
        raise MlflowLoggingError(
            f"Required keys for this dictionary include {', '.join(required_keys)}."
        )


def get_mlflow_client(
    workspace_kwargs: dict = {}, service_principal_kwargs: dict = {}
) -> MlflowClient:
    """
    Set remote tracking URI for mlflow to AzureML workspace

    Parameters
    ----------
    workspace_kwargs: dict
        AzureML Workspace configuration to use for remote MLFlow tracking. An
        empty dict will result in local logging by the MlflowClient.
        Example::

            `{
                 "subscription_id":<value>,
                 "resource_group":<value>,
                 "workspace_name":<value>
             }`
    service_principal_kwargs: dict
        AzureML ServicePrincipalAuthentication keyword arguments. An empty dict
        will result in interactive authentication.
        Example::

            `{
                 "tenant_id":<value>,
                 "service_principal_id":<value>,
                 "service_principal_password":<value>
             }`

    Returns
    -------
    client: mlflow.tracking.MlflowClient
        Client with tracking uri set to AzureML if configured.
    """
    logger.info("Creating MLflow tracking client.")

    tracking_uri = None

    # Get AzureML tracking_uri if using Azure as backend
    if workspace_kwargs:
        required_keys = ["subscription_id", "resource_group", "workspace_name"]
        _validate_dict(workspace_kwargs, required_keys)

        msg = "Configuring AzureML backend with {auth_type} authentication."
        if service_principal_kwargs:
            required_keys = [
                "tenant_id",
                "service_principal_id",
                "service_principal_password",
            ]
            _validate_dict(service_principal_kwargs, required_keys)

            logger.info(msg.format(auth_type="ServicePrincipalAuthentication"))
            workspace_kwargs["auth"] = ServicePrincipalAuthentication(
                **service_principal_kwargs
            )
        else:
            logger.info(msg.format(auth_type="interactive"))
            workspace_kwargs["auth"] = InteractiveLoginAuthentication(force=True)

        tracking_uri = Workspace(**workspace_kwargs).get_mlflow_tracking_uri()

    mlflow.set_tracking_uri(tracking_uri)

    return MlflowClient(tracking_uri)


def get_run_id(client: MlflowClient, experiment_name: str, model_key: str) -> str:
    """
    Get an existing or create a new run for the given model_key and experiment_name.

    The model key corresponds to a unique configuration of the model. The corresponding
    run must be manually stopped using the `mlflow.tracking.MlflowClient.set_terminated`
    method.

    Parameters
    ----------
    client: mlflow.tracking.MlflowClient
        Client with tracking uri set to AzureML if configured.
    experiment_name: str
        Name of experiment to log to.
    model_key: str
        Unique ID of model configuration.

    Returns
    -------
    run_id: str
        Unique ID of MLflow run to log to.
    """
    experiment = client.get_experiment_by_name(experiment_name)

    experiment_id = (
        getattr(experiment, "experiment_id")
        if experiment
        else client.create_experiment(experiment_name)
    )
    return client.create_run(experiment_id, tags={"model_key": model_key}).info.run_id


def _datetime_to_ms_since_epoch(dt: datetime) -> int:
    """
    Convert datetime to milliseconds since Unix epoch (UTC)

    Parameters
    ----------
    dt: datetime.datetime
        Timestamp to convert (can be timezone aware or naive).

    Returns
    -------
    dt: int
        Timestamp as milliseconds since Unix epoch

    Example
    -------
    >>> dt = datetime(1970, 1, 1, 0, 0)
    >>> _datetime_to_ms_since_epoch(dt)
    0
    """
    epoch = datetime.utcfromtimestamp(0).replace(tzinfo=dt.tzinfo)
    dt = dt.astimezone(UTC) if dt.tzinfo else dt.replace(tzinfo=dt.tzinfo)
    return round((dt - epoch).total_seconds() * 1000.0)


def epoch_now() -> int:
    """
    Get current timestamp in UTC as milliseconds since Unix epoch.

    Returns
    -------
    now: int
        Milliseconds since Unix epoch.
    """
    return _datetime_to_ms_since_epoch(datetime.now(tz=UTC))


def get_machine_log_items(machine: Machine) -> Tuple[List[Metric], List[Param]]:
    """
    Create flat lists of MLflow logging entities from multilevel dictionary

    For more information, see the mlflow docs:
    https://www.mlflow.org/docs/latest/python_api/mlflow.tracking.html#mlflow.tracking.MlflowClient.log_batch

    Parameters
    ----------
    machine: Machine

    Returns
    -------
    metrics: List[Metric]
        List of MLFlow Metric objects to log.
    params: List[Param]
        List of MLFlow Param objects to log.
    """

    metrics: List[Metric] = list()
    build_metadata = machine.metadata.build_metadata

    # Project/machine parameters
    keys = ["project_name", "name"]
    params = [Param(attr, getattr(machine, attr)) for attr in keys]

    # Dataset parameters
    dataset_keys = [
        "train_start_date",
        "train_end_date",
        "resolution",
        "row_filter",
        "row_filter_buffer_size",
    ]
    params.extend(Param(k, str(getattr(machine.dataset, k))) for k in dataset_keys)

    # Model parameters
    model_keys = ["model_creation_date", "model_builder_version", "model_offset"]
    params.extend(Param(k, str(getattr(build_metadata.model, k))) for k in model_keys)

    # Parse cross-validation split metadata
    splits = build_metadata.model.cross_validation.splits
    params.extend(Param(k, str(v)) for k, v in splits.items())

    # Parse cross-validation metrics

    tag_list = machine.normalize_sensor_tags(machine.dataset.tag_list)
    scores = build_metadata.model.cross_validation.scores

    keys = sorted(list(scores.keys()))
    subkeys = ["mean", "max", "min", "std"]

    n_folds = len(scores[keys[0]]) - len(subkeys)
    for k in keys:
        # Skip per tag data, produces too many params for MLflow
        if any([extract_tag_name(t) in k for t in tag_list]):
            continue

        # Summary stats per metric
        for sk in subkeys:
            metrics.append(Metric(f"{k}-{sk}", scores[k][f"fold-{sk}"], epoch_now(), 0))
        # Append value for each fold with increasing steps
        metrics.extend(
            Metric(k, scores[k][f"fold-{i+1}"], epoch_now(), i) for i in range(n_folds)
        )

    # Parse fit metrics
    try:
        history_metrics = build_metadata.model.model_meta["history"]
        meta_parms = build_metadata.model.model_meta["history"]["params"]
    except KeyError:
        logger.debug(
            "Key 'build-metadata.model.history.params' not found found in metadata."
        )
    else:
        metrics.extend(
            Metric(k, float(getattr(build_metadata.model, k)), epoch_now(), 0)
            for k in ["model_training_duration_sec"]
        )
        for m in (x for x in history_metrics if x != "params"):
            data = build_metadata.model.model_meta["history"][m]
            metrics.extend(
                Metric(m, float(x), timestamp=epoch_now(), step=i)
                for i, x in enumerate(data)
            )
        params.extend(Param(k, str(meta_parms[k])) for k in meta_parms)

    return metrics, params


def batch_log_items(
    metrics: List[Metric],
    params: List[Param],
    n_max_metrics: int = 200,
    n_max_params: int = 100,
) -> List[Dict[str, Union[Metric, Param]]]:
    """
    Split metrics, params and tags to batches that satisfy limits imposed by MLFlow and AzureML

    NOTE: The default maximum number of metrics are and parameters set here are
    those set by AzureML as per today, 18 February 2020.

    Also, there the 1mb request size is not evaluated here, as doing this
    should not be necessary and is not addressable in a succint way. MLflow
    also has a limit of 1000 log items per request, but reaching this is not
    possible with AzureML's current limit on metrics.

    Parameters
    ----------
    metrics: List[Metric]
        List of MLFlow Metric objects to log.
    params: List[Param]
        List of MLFlow Param objects to log.
    n_max_metrics:int
        Limit to number of metrics AzureML allows per batch log request payload.
    n_max_params:int
        Limit to number of params MLFlow allows per batch log request payload.

    Returns
    -------
    log_batches: List[Dict[str, Union[Metric, Param]]]
        List of MlflowClinet.log_batch keyworkd arguments, split to quatnitites
        that respect limits present for MLFlow and AzureML.
    """

    def _calc_n_batches(n: int, n_max: int):
        """
        Calculate the number of batches required to log n items with batches of n_max
        """
        return (n // n_max) + (1 if (n % n_max) else 0)

    n_batches = max(
        _calc_n_batches(len(metrics), n_max_metrics),
        _calc_n_batches(len(params), n_max_params),
    )

    i = 0
    j = 0
    log_batches = list()
    for _ in range(n_batches):
        log_batches.append(
            {
                "metrics": metrics[i : i + n_max_metrics],
                "params": params[j : j + n_max_params],
            }
        )
        i += n_max_metrics
        j += n_max_params

    return log_batches


def get_kwargs_from_secret(name: str, keys: List[str]) -> dict:
    """
    Get keyword arguments dictionary from secrets environment variable

    Parameters
    ----------
    name: str
        Name of the environment variable whose content is a colon separated
        list of secrets.

    Returns
    -------
    kwargs: dict
        Dictionary of keyword arguments parsed from environment variable.
    """
    secret_str = os.getenv(name)

    if secret_str is None:
        raise MlflowLoggingError(f"The value for env var '{name}' must not be `None`.")

    if secret_str:
        elements = secret_str.split(":")
        if len(elements) != len(keys):
            raise MlflowLoggingError(
                "`keys` len {len(keys)} must be of equal length with env var {name} elements {len(elements)}."
            )
        kwargs = {key: elements[i] for i, key in enumerate(keys)}
    else:
        kwargs = {}

    return kwargs


def get_workspace_kwargs() -> dict:
    """Get AzureML keyword arguments from environment

    The name of this environment variable is set in the Argo workflow template,
    and its value should be in the format:
    `<subscription_id>:<resource_group>:<workspace_name>`.

    Returns
    -------
    workspace_kwargs: dict
        AzureML Workspace configuration to use for remote MLFlow tracking. See
        :func:`gordo.builder.mlflow_utils.get_mlflow_client`.
    """
    return get_kwargs_from_secret(
        "AZUREML_WORKSPACE_STR", ["subscription_id", "resource_group", "workspace_name"]
    )


def get_spauth_kwargs() -> dict:
    """Get AzureML keyword arguments from environment

    The name of this environment variable is set in the Argo workflow template,
    and its value should be in the format:
    `<tenant_id>:<service_principal_id>:<service_principal_password>`

    Returns
    -------
    service_principal_kwargs: dict
        AzureML ServicePrincipalAuthentication keyword arguments. See
        :func:`gordo.builder.mlflow_utils.get_mlflow_client`
    """
    return get_kwargs_from_secret(
        "DL_SERVICE_AUTH_STR",
        ["tenant_id", "service_principal_id", "service_principal_password"],
    )


@contextmanager
def mlflow_context(
    name: str,
    model_key: str = uuid4().hex,
    workspace_kwargs: dict = {},
    service_principal_kwargs: dict = {},
):
    """
    Generate MLflow logger function with either a local or AzureML backend

    Parameters
    ----------
    name: str
        The name of the log group to log to (e.g. a model name).
    model_key: str
        Unique ID of logging run.
    workspace_kwargs: dict
        AzureML Workspace configuration to use for remote MLFlow tracking. See
        :func:`gordo.builder.mlflow_utils.get_mlflow_client`.
    service_principal_kwargs: dict
        AzureML ServicePrincipalAuthentication keyword arguments. See
        :func:`gordo.builder.mlflow_utils.get_mlflow_client`

    Example
    -------
    >>> with tempfile.TemporaryDirectory as tmp_dir:
    ...     mlflow.set_tracking_uri(f"file:{tmp_dir}")
    ...     with mlflow_context("log_group", "unique_key", {}, {}) as (mlflow_client, run_id):
    ...         log_machine(machine) # doctest: +SKIP
    """
    mlflow_client = get_mlflow_client(workspace_kwargs, service_principal_kwargs)
    run_id = get_run_id(mlflow_client, experiment_name=name, model_key=model_key)

    logger.info(
        f"MLflow client configured to use {'AzureML' if workspace_kwargs else 'local backend'}"
    )

    yield mlflow_client, run_id

    mlflow_client.set_terminated(run_id)


def log_machine(mlflow_client: MlflowClient, run_id: str, machine: Machine):
    """
    Send logs to configured MLflow backend

    Parameters
    ----------
    mlflow_client: MlflowClient
        Client instance to call logging methods from.
    run_id: str
        Unique ID off MLflow Run to log to.
    machine: Machine
        Machine to log with MlflowClient.
    """
    # Log machine metrics and params
    for batch_kwargs in batch_log_items(*get_machine_log_items(machine)):
        mlflow_client.log_batch(run_id, **batch_kwargs)

    # Send configs as JSON artifacts
    try:
        with tempfile.TemporaryDirectory(dir="./") as tmp_dir:
            fp = os.path.join(tmp_dir, "metadata.json")
            with open(fp, "w") as fh:
                json.dump(machine.to_dict(), fh, cls=MachineJSONEncoder)
            mlflow_client.log_artifacts(run_id=run_id, local_dir=tmp_dir)
    # Map to MlflowLoggingError for coding errors in the model builder
    except Exception as e:
        raise MlflowLoggingError(e)


class MlFlowReporter(BaseReporter):
    @capture_args
    def __init__(
        self,
        *args,
        model_builder_class: Optional[Union[str, Type[ModelBuilder]]] = None,
        **kwargs,
    ):
        if type(model_builder_class) is str:
            model_builder_class = create_model_builder(model_builder_class)
        if model_builder_class is None:
            model_builder_class = ModelBuilder
        self.model_builder_class = cast(Type[ModelBuilder], model_builder_class)

    def report(self, machine: Machine):
        workspace_kwargs = get_workspace_kwargs()
        service_principal_kwargs = get_spauth_kwargs()
        # TODO something better here
        model_builder = self.model_builder_class(machine)
        cache_key = model_builder.calculate_cache_key(machine)

        with mlflow_context(
            machine.name, cache_key, workspace_kwargs, service_principal_kwargs
        ) as (mlflow_client, run_id):
            log_machine(mlflow_client, run_id, machine)



import pytest
import yaml
import re

from gordo.machine.loader import (
    load_globals_config,
    load_machine_config,
    load_model_config,
    MachineConfigException,
)


def test_load_globals_config():
    expected_config = {
        "dataset": {
            "tags": ["tag1", "tag2"],
            "train_end_date": "2022-01-15T00:00:00+00:00",
            "train_start_date": "2021-12-25T00:00:00+00:00",
        },
        "model": {
            "gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector": {
                "base_estimator": {
                    "sklearn.pipeline.Pipeline": {
                        "steps": [
                            "sklearn.preprocessing.MinMaxScaler",
                            {
                                "gordo.machine.model.models.KerasAutoEncoder": {
                                    "kind": "feedforward_hourglass"
                                }
                            },
                        ]
                    }
                }
            }
        },
        "runtime": {"builder": {"resources": {"requests": {"memory": 1000}}}},
        "metadata": {"key1": "value1"},
        "evaluation": {"cv_mode": "no_cv"},
    }
    config_content = """
    dataset: |
      tags:
      - tag1
      - tag2
      train_end_date: '2022-01-15T00:00:00+00:00'
      train_start_date: '2021-12-25T00:00:00+00:00'
    model: |
      gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
        base_estimator:
          sklearn.pipeline.Pipeline:
            steps:
              - sklearn.preprocessing.MinMaxScaler
              - gordo.machine.model.models.KerasAutoEncoder:
                  kind: feedforward_hourglass
    runtime: |
      builder:
        resources:
          requests:
            memory: 1000
    metadata: |
      key1: value1
    evaluation: |
      cv_mode: no_cv
    """
    global_config = load_globals_config(
        yaml.safe_load(config_content), "spec.config.globals"
    )
    assert global_config == expected_config
    # Test for legacy config logic
    config_content = """
    dataset:
      tags:
      - tag1
      - tag2
      train_end_date: '2022-01-15T00:00:00+00:00'
      train_start_date: '2021-12-25T00:00:00+00:00'
    model: |
      gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
        base_estimator:
          sklearn.pipeline.Pipeline:
            steps:
              - sklearn.preprocessing.MinMaxScaler
              - gordo.machine.model.models.KerasAutoEncoder:
                  kind: feedforward_hourglass
    runtime:
      builder:
        resources:
          requests:
            memory: 1000
    metadata:
      key1: value1
    evaluation:
      cv_mode: no_cv
    """
    global_config = load_globals_config(
        yaml.safe_load(config_content), "spec.config.globals"
    )
    config_content = """
    dataset:
      tags:
      - tag1
      - tag2
      train_end_date: '2022-01-15T00:00:00+00:00'
      train_start_date: '2021-12-25T00:00:00+00:00'
    """
    global_config = load_globals_config(
        yaml.safe_load(config_content), "spec.config.globals"
    )
    assert global_config == {
        "dataset": {
            "tags": ["tag1", "tag2"],
            "train_end_date": "2022-01-15T00:00:00+00:00",
            "train_start_date": "2021-12-25T00:00:00+00:00",
        }
    }
    config_content = """
    dataset: |
      tags:
          - tag1
      - tag2
    """
    with pytest.raises(
        MachineConfigException,
        match=re.escape(r"Error loading YAML from 'spec.config.globals.dataset'"),
    ):
        load_globals_config(yaml.safe_load(config_content), "spec.config.globals")


def test_load_machine_config():
    config_content = """
    name: 'model1'
    dataset: |
      tags:
      - tag1
      - tag2
      train_end_date: '2022-01-15T00:00:00+00:00'
      train_start_date: '2021-12-25T00:00:00+00:00'
    """
    machine_config = load_machine_config(
        yaml.safe_load(config_content), "spec.config.machines[0]"
    )
    assert machine_config == {
        "name": "model1",
        "dataset": {
            "tags": ["tag1", "tag2"],
            "train_end_date": "2022-01-15T00:00:00+00:00",
            "train_start_date": "2021-12-25T00:00:00+00:00",
        },
    }
    config_content = """
    dataset: |
      tags:
      - tag1
      - tag2
      train_start_date: '2021-12-25T00:00:00+00:00'
    """
    with pytest.raises(
        MachineConfigException,
        match=re.escape(r"'spec.config.machines[0].name' is empty"),
    ):
        load_machine_config(yaml.safe_load(config_content), "spec.config.machines[0]")


def test_load_model_config():
    config_content = """
    name: 'model1'
    project_name: 'project1'
    dataset: |
      tags:
      - tag1
      - tag2
      train_end_date: '2022-01-15T00:00:00+00:00'
      train_start_date: '2021-12-25T00:00:00+00:00'
    """
    model_config = load_model_config(yaml.safe_load(config_content), "spec.config")
    assert model_config == {
        "name": "model1",
        "project_name": "project1",
        "dataset": {
            "tags": ["tag1", "tag2"],
            "train_end_date": "2022-01-15T00:00:00+00:00",
            "train_start_date": "2021-12-25T00:00:00+00:00",
        },
    }
    config_content = """
    name: 'model1'
    dataset: |
      tags:
      - tag1
      - tag2
      train_start_date: '2021-12-25T00:00:00+00:00'
    """
    with pytest.raises(
        MachineConfigException, match=re.escape(r"'spec.config.project_name' is empty")
    ):
        load_model_config(yaml.safe_load(config_content), "spec.config")



import os

from pathlib import Path
from setuptools import setup, find_packages


setup_requirements = ["pytest-runner", "setuptools_scm"]
on_rtd = os.environ.get("READTHEDOCS") == "True"


def requirements(fp: str):
    with open(os.path.join(os.path.dirname(__file__), "requirements", fp)) as f:
        return [
            r.strip()
            for r in f.readlines()
            if r.strip() and not r.startswith("#") and not r.startswith("-")
        ]


extras_require = {
    "docs": requirements("docs_requirements.in"),
    "mlflow": requirements("mlflow_requirements.in"),
    "postgres": requirements("postgres_requirements.in"),
    "tests": requirements("test_requirements.txt"),
}
extras_require["full"] = extras_require["mlflow"] + extras_require["postgres"]

install_requires = requirements("requirements.in")  # Allow flexible deps for install

# Read the docs have quite low memory limits on their build servers, so low
# that pip crashes when downloading tensorflow. So we must remove it from the install
# requirements, and set up autodoc_mock_imports in docs/conf.py
if on_rtd:
    install_requires = [req for req in install_requires if "tensorflow" not in req]

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    author="Equinor ASA",
    author_email="fg_gpl@equinor.com",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="Train and build models for Argo / Kubernetes",
    long_description=long_description,
    long_description_content_type='text/markdown',
    entry_points={"console_scripts": ["gordo=gordo.cli:gordo"]},
    install_requires=install_requires,
    license="AGPLv3",
    name="gordo",
    packages=find_packages(),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=extras_require["tests"],
    extras_require=extras_require,
    url="https://github.com/equinor/gordo",
    use_scm_version={"write_to": "gordo/_version.py", "relative_to": __file__},
    zip_safe=True,
    package_data={
        "": [
            "gordo/workflow/workflow_generator/resources/argo-workflow.yml.template",
            "gordo/machine/dataset/data_provider/resources/assets_config.yaml",
        ]
    },
    include_package_data=True,
)



import pytest

from gordo.builder.build_model import ModelBuilder
from gordo.builder.utils import create_model_builder


def test_create_model_builder_success():
    cls = create_model_builder(None)
    assert cls is ModelBuilder
    cls = create_model_builder("gordo.builder.build_model.ModelBuilder")
    assert cls is ModelBuilder


def test_create_model_builder_failed():
    with pytest.raises(ImportError):
        create_model_builder("wrong.import.WrongClass")
    with pytest.raises(
        ValueError,
        match='"Machine" class located in "gordo.machine.Machine" should be subclass of "ModelBuilder"',
    ):
        create_model_builder("gordo.machine.Machine")



# -*- coding: utf-8 -*-

"""
CLI interfaces
"""

import logging
import sys
import traceback
import jinja2
import yaml
import click

from gordo_core.data_providers.providers import NoSuitableDataProviderError
from gordo_core.sensor_tag import SensorTagNormalizationError
from gordo_core.exceptions import ConfigException, InsufficientDataError
from gunicorn.glogging import Logger
from typing import Tuple, List, Any, cast
from gordo.builder.utils import create_model_builder

from gordo import serializer
from gordo.server import server
from gordo import __version__
from gordo.machine import Machine, load_model_config
from gordo.cli.workflow_generator import workflow_cli
from gordo.cli.custom_types import key_value_par, HostIP
from gordo.reporters.exceptions import ReporterException

from .exceptions_reporter import ReportLevel, ExceptionsReporter

_exceptions_reporter = ExceptionsReporter(
    (
        (Exception, 1),
        (ValueError, 2),
        (PermissionError, 20),
        (FileNotFoundError, 30),
        (SensorTagNormalizationError, 60),
        (NoSuitableDataProviderError, 70),
        (InsufficientDataError, 80),
        (ImportError, 85),
        (ReporterException, 90),
        (ConfigException, 100),
    )
)

logger = logging.getLogger(__name__)


@click.group("gordo")
@click.version_option(version=__version__, message=__version__)
@click.option(
    "--log-level",
    type=str,
    default="INFO",
    help="Run workflow with custom log-level.",
    envvar="GORDO_LOG_LEVEL",
)
@click.pass_context
def gordo(gordo_ctx: click.Context, **ctx):
    """
    The main entry point for the CLI interface
    """
    # Set log level, defaulting to INFO
    logging.basicConfig(
        level=getattr(logging, str(gordo_ctx.params.get("log_level")).upper()),
        format="[%(asctime)s] %(levelname)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s",
    )

    # Set matplotlib log level to INFO to avoid noise
    logging.getLogger("matplotlib").setLevel(logging.INFO)

    gordo_ctx.obj = gordo_ctx.params


@click.command()
@click.argument("machine-config", envvar="MACHINE", type=yaml.safe_load)
@click.argument("output-dir", default="/data", envvar="OUTPUT_DIR")
@click.option(
    "--model-register-dir",
    default=None,
    envvar="MODEL_REGISTER_DIR",
    type=click.Path(
        exists=False, file_okay=False, dir_okay=True, writable=True, readable=True
    ),
)
@click.option(
    "--model-builder-class",
    help="ModelBuilder class import path. "
    "This should be a subclass of gordo.builder.build_model.ModelBuilder",
    envvar="MODEL_BUILDER_CLASS",
)
@click.option(
    "--print-cv-scores", help="Prints CV scores to stdout", is_flag=True, default=False
)
@click.option(
    "--model-parameter",
    type=key_value_par,
    multiple=True,
    default=(),
    help="Key-Value pair for a model parameter and its value, may use this option "
    "multiple times. Separate key,valye by a comma. ie: --model-parameter key,val "
    "--model-parameter some_key,some_value",
)
@click.option(
    "--exceptions-reporter-file",
    envvar="EXCEPTIONS_REPORTER_FILE",
    help="JSON output file for exception information",
)
@click.option(
    "--exceptions-report-level",
    type=click.Choice(ReportLevel.get_names(), case_sensitive=False),
    default=ReportLevel.MESSAGE.name,
    envvar="EXCEPTIONS_REPORT_LEVEL",
    help="Details level for exception reporting",
)
def build(
    machine_config: dict,
    output_dir: str,
    model_register_dir: click.Path,
    model_builder_class: str,
    print_cv_scores: bool,
    model_parameter: List[Tuple[str, Any]],
    exceptions_reporter_file: str,
    exceptions_report_level: str,
):
    """
    Build a model and deposit it into 'output_dir' given the appropriate config
    settings.

    \b
    Parameters
    ----------
    machine_config: dict
        A dict loadable by :class:`gordo.machine.Machine.from_config`
    output_dir: str
        Directory to save model & metadata to.
    model_register_dir: path
        Path to a directory which will index existing models and their locations, used
        for re-using old models instead of rebuilding them. If omitted then always
        rebuild
    model_builder_class: str
    print_cv_scores: bool
        Print cross validation scores to stdout
    model_parameter: List[Tuple[str, Any]
        List of model key-values, wheres the values will be injected into the model
        config wherever there is a jinja variable with the key.
    exceptions_reporter_file: str
        JSON output file for exception information
    exceptions_report_level: str
        Details level for exception reporting
    """

    try:
        if model_parameter and isinstance(machine_config["model"], str):
            parameters = dict(model_parameter)  # convert lib of tuples to dict
            machine_config["model"] = expand_model(machine_config["model"], parameters)

        machine: Machine = Machine.from_config(
            cast(dict[str, Any], load_model_config(machine_config)),
            project_name=machine_config["project_name"],
        )

        logger.info(f"Building, output will be at: {output_dir}")
        logger.info(f"Register dir: {model_register_dir}")

        # Convert the config into a pipeline, and back into definition to ensure
        # all default parameters are part of the config.
        logger.debug("Ensuring the passed model config is fully expanded.")
        machine.model = serializer.into_definition(
            serializer.from_definition(machine.model)
        )
        logger.info("Fully expanded model config: {machine.model}")

        cls = create_model_builder(model_builder_class)
        builder = cls(machine=machine)

        _, machine_out = builder.build(output_dir, model_register_dir)  # type: ignore

        logger.debug("Reporting built machine.")
        machine_out.report()
        logger.debug("Finished reporting.")

        if "err" in machine.name:
            raise FileNotFoundError("undefined_file.parquet")

        if print_cv_scores:
            for score in get_all_score_strings(machine_out):
                print(score)

    except Exception:
        traceback.print_exc()
        exc_type, exc_value, exc_traceback = sys.exc_info()

        exit_code = _exceptions_reporter.exception_exit_code(exc_type)
        if exceptions_reporter_file:
            _exceptions_reporter.safe_report(
                cast(
                    ReportLevel,
                    ReportLevel.get_by_name(
                        exceptions_report_level, ReportLevel.EXIT_CODE
                    ),
                ),
                exc_type,
                exc_value,
                exc_traceback,
                exceptions_reporter_file,
                max_message_len=2024 - 500,
            )
        sys.exit(exit_code)
    else:
        return 0


def expand_model(model_config: str, model_parameters: dict):
    """
    Expands the jinja template which is the model using the variables in
    `model_parameters`

    Parameters
    ----------
    model_config: str
        Jinja template which when expanded becomes a valid model config json.
    model_parameters:
        Parameters for the model config.

    Raises
    ------
    ValueError
        If an undefined variable is used in the model_config.

    Returns
    -------
    str
        The model config with variables expanded

    """
    try:
        model_template = jinja2.Environment(
            loader=jinja2.BaseLoader(), undefined=jinja2.StrictUndefined
        ).from_string(model_config)
        model_config = model_template.render(**model_parameters)
    except jinja2.exceptions.UndefinedError as e:
        raise ValueError("Model parameter missing value!") from e
    logger.info(f"Expanded model config: {model_config}")
    return yaml.safe_load(model_config)


def get_all_score_strings(machine):
    """Given metadata from the model builder this function returns a list of
    strings of the following format:
    {metric_name}-{tag_name}_{fold-fold-number} = {score_val}.  This computes the score for the given tag and
    cross validation split.
    {metric_name}-{tag_name}_{fold-aggregation} = {score_val}. This computes the score for the given tag and aggregates
    the score over all cross validation splits (aggregations currently used are mean, std, min and max)
    {metric_name}_{fold-fold-number} = {score_val}.  This computes the score aggregate across all tags (uses sklearn's default
    aggregation method) for a given cross validation split.
    {metric_name}_{fold-aggregation} = {score_val}.  This computes the score aggregate across all tags (uses sklearn's default
    aggregation method) and cross validation splits (aggregations currently used are mean, std, min and max).

    for katib to pick up.

    Current metric names supported are sklearn score functions: 'r2_score', 'explained_variance_score',
    'mean_squared_error' and 'mean_absolute_error'.  The underscores in such score names are replaced by '-'.

    All spaces in the tag name are also replaced by '-'.

    Parameters
    ----------
    machine : Machine
    """
    all_scores = []
    for (
        metric_name,
        scores,
    ) in machine.metadata.build_metadata.model.cross_validation.scores.items():
        metric_name = metric_name.replace(" ", "-")
        for score_name, score_val in scores.items():
            score_name = score_name.replace(" ", "-")
            all_scores.append(f"{metric_name}_{score_name}={score_val}")
    return all_scores


@click.command("run-server")
@click.option(
    "--host",
    type=HostIP(),
    help="The host to run the server on.",
    default="0.0.0.0",
    envvar="GORDO_SERVER_HOST",
    show_default=True,
)
@click.option(
    "--port",
    type=click.IntRange(1, 65535),
    help="The port to run the server on.",
    default=5555,
    envvar="GORDO_SERVER_PORT",
    show_default=True,
)
@click.option(
    "--workers",
    type=click.IntRange(1, 4),
    help="The number of worker processes for handling requests.",
    default=2,
    envvar="GORDO_SERVER_WORKERS",
    show_default=True,
)
@click.option(
    "--worker-connections",
    type=click.IntRange(1, 4000),
    help="The maximum number of simultaneous clients per worker process.",
    default=50,
    envvar="GORDO_SERVER_WORKER_CONNECTIONS",
    show_default=True,
)
@click.option(
    "--threads",
    type=int,
    help="The number of worker threads for handling requests."
    "This argument only has affects with --worker-class=gthread. "
    "Default value is 8 (4 x $(NUM_CORES))",
    default=8,
    envvar="GORDO_SERVER_THREADS",
)
@click.option(
    "--worker-class",
    help="The type of workers to use.",
    default="gthread",
    envvar="GORDO_SERVER_WORKER_CLASS",
    show_default=True,
)
@click.option(
    "--log-level",
    type=click.Choice(Logger.LOG_LEVELS.keys()),
    help="The log level for the server.",
    default="debug",
    envvar="GORDO_SERVER_LOG_LEVEL",
    show_default=True,
)
@click.option(
    "--server-app",
    help="The application to run",
    default="gordo.server.server:build_app()",
    envvar="GORDO_SERVER_APP",
    show_default=True,
)
@click.option(
    "--with-prometheus-config",
    help="Run with custom config for prometheus",
    is_flag=True,
)
def run_server_cli(
    host,
    port,
    workers,
    worker_connections,
    threads,
    worker_class,
    log_level,
    server_app,
    with_prometheus_config,
):
    """
    Run the gordo server app with Gunicorn
    """
    config_module = None
    if with_prometheus_config:
        config_module = "gordo.server.prometheus.gunicorn_config"
    server.run_server(
        host,
        port,
        workers,
        log_level.lower(),
        config_module=config_module,
        worker_connections=worker_connections,
        threads=threads,
        worker_class=worker_class,
        server_app=server_app,
    )


gordo.add_command(workflow_cli)
gordo.add_command(build)
gordo.add_command(run_server_cli)


if __name__ == "__main__":
    gordo()



# -*- coding: utf-8 -*-
import os
import sys


def test_formatting_black():
    project_path = os.path.join(os.path.dirname(__file__), "..")
    gordo_path = os.path.join(project_path, "gordo")
    tests_path = os.path.join(project_path, "tests")
    cmd = [
        sys.executable,
        "-m",
        "black",
        "--check",
        "-v",
        gordo_path,
        tests_path,
        "--exclude",
        r".*_version.py",
    ]
    exit_code = os.system(" ".join(cmd))
    assert exit_code == 0



# -*- coding: utf-8 -*-

import typing
import functools
import logging
from typing import Optional, Union, List, Iterable, Sized, cast
from datetime import timedelta, datetime

import numpy as np
import pandas as pd
from sklearn.base import TransformerMixin

from gordo_core.sensor_tag import SensorTag

logger = logging.getLogger(__name__)


def metric_wrapper(metric, scaler: Optional[TransformerMixin] = None):
    """
    Ensures that a given metric works properly when the model itself returns
    a y which is shorter than the target y, and allows scaling the data
    before applying the metrics.


    Parameters
    ----------
    metric
        Metric which must accept y_true and y_pred of the same length
    scaler :  Optional[TransformerMixin]
        Transformer which will be applied on y and y_pred before the metrics is
        calculated. Must have method `transform`, so for most scalers it must already
        be fitted on `y`.
    """

    @functools.wraps(metric)
    def _wrapper(y_true, y_pred, *args, **kwargs):
        if scaler:
            logger.debug(
                "Transformer provided to metrics wrapper, scaling y and y_pred before "
                "passing to metrics"
            )
            y_true = scaler.transform(y_true)
            y_pred = scaler.transform(y_pred)
        return metric(y_true[-len(y_pred) :], y_pred, *args, **kwargs)

    return _wrapper


def make_base_dataframe(
    tags: typing.Union[typing.List[SensorTag], typing.List[str]],
    model_input: np.ndarray,
    model_output: np.ndarray,
    target_tag_list: Optional[Union[List[SensorTag], List[str]]] = None,
    index: typing.Optional[Union[np.ndarray, pd.Index]] = None,
    frequency: typing.Optional[timedelta] = None,
) -> pd.DataFrame:
    """
    Construct a dataframe which has a MultiIndex column consisting of top level keys
    'model-input' and 'model-output'. Takes care of aligning model output if different
    than model input lengths, as setting column names based on passed tags and target_tag_list.

    Parameters
    ----------
    tags: List[Union[str, SensorTag]]
        Tags which will be assigned to ``model-input`` and/or ``model-output`` if
        the shapes match.
    model_input: np.ndarray
        Original input given to the model
    model_output: np.ndarray
        Raw model output
    target_tag_list: Optional[Union[List[SensorTag], List[str]]]
        Tags to be assigned to ``model-output`` if not assigned but model output matches
        model input, ``tags`` will be used.
    index: Optional[np.ndarray]
        The index which should be assigned to the resulting dataframe, will be clipped
        to the length of ``model_output``, should the model output less than its input.
    frequency: Optional[datetime.timedelta]
        The spacing of the time between points.

    Returns
    -------
    pd.DataFrame
    """

    # Set target_tag_list to default to tags if not specified.
    target_tag_list = target_tag_list if target_tag_list is not None else tags

    # match length of output, and ensure we're working with numpy arrays, not pandas.
    model_input = getattr(model_input, "values", model_input)[-len(model_output) :, :]
    model_output = getattr(model_output, "values", model_output)

    names_n_values = (("model-input", model_input), ("model-output", model_output))

    # Define the index which all series/dataframes will share
    normalized_index = (
        index[-len(model_output) :]
        if index is not None
        else list(range(len(model_output)))
    )

    # Series to hold the start times for each point or just 'None' values
    series_values: Union[np.ndarray, Iterable]
    if isinstance(normalized_index, pd.DatetimeIndex):
        series_values = normalized_index
    else:
        series_values = []
        if (
            isinstance(normalized_index, pd.Index) and not normalized_index.empty
        ) or normalized_index:
            series_values = (None for _ in range(len(cast(Sized, normalized_index))))
    start_series = pd.Series(series_values, index=normalized_index)

    # Calculate the end times if possible, or also all 'None's
    end_series = start_series.map(
        lambda start: (start + frequency).isoformat()
        if isinstance(start, datetime) and frequency is not None
        else None
    )

    # Convert to isoformatted string for JSON serialization.
    start_series = start_series.map(
        lambda start: start.isoformat() if hasattr(start, "isoformat") else None
    )

    # The resulting DF will be multiindex, so we define and initialize it here
    # with the start and end times from above.
    columns = pd.MultiIndex.from_product((("start", "end"), ("",)))
    data: pd.DataFrame = pd.DataFrame(
        {("start", ""): start_series, ("end", ""): end_series},
        columns=columns,
        index=normalized_index,
    )

    # Begin looping over the model-input and model-output; mapping them into
    # the multiindex column dataframe, and naming their second level labels as needed.
    name: str
    values: np.ndarray
    for (name, values) in filter(lambda nv: nv[1] is not None, names_n_values):

        _tags = tags if name == "model-input" else target_tag_list

        # Create the second level of column names, either as the tag names
        # or simple range of numbers
        if values.shape[1] == len(_tags):
            # map(...) to satisfy mypy to match second possible outcome
            second_lvl_names = map(
                str, (tag.name if isinstance(tag, SensorTag) else tag for tag in _tags)
            )
        else:
            second_lvl_names = map(str, range(values.shape[1]))

        # Columns will be multi level with the title of the output on top
        # and specific names below, ie. ('model-output', 'tag-0') as a column
        columns = pd.MultiIndex.from_tuples(
            (name, sub_name) for sub_name in second_lvl_names
        )

        # Pass valudes, offsetting any differences in length compared to index, as set by model-output size
        other = pd.DataFrame(
            values[-len(model_output) :], columns=columns, index=normalized_index
        )
        data = data.join(other)

    return data



from .from_definition import from_definition, load_params_from_definition
from .into_definition import into_definition, load_definition_from_params
from .serializer import (
    dump,
    dumps,
    load,
    loads,
    load_metadata,
    metadata_path,
    load_info,
)
